@@ -1,13 +1,19 @@
 # coding=utf-8
 import logging
+
+from requests.exceptions import HTTPError
+
 from .rest_client import AtlassianRestAPI
 
 log = logging.getLogger(__name__)
 
 
 class Bamboo(AtlassianRestAPI):
+
+    """ Private methods """
+
     def _get_generator(self, path, elements_key='results', element_key='result', data=None, flags=None,
-                       params=None, headers=None):
+                       params=None, headers=None, max_results=None):
         """
         Generic method to return a generator with the results returned from Bamboo. It is intended to work for
         responses in the form:
@@ -27,22 +33,32 @@ class Bamboo(AtlassianRestAPI):
         :param path: URI for the resource
         :return: generator with the contents of response[elements_key][element_key]
         """
-        size = 1
         start_index = 0
-        while size:
-            params['start-index'] = start_index
-            response = self.get(path, data, flags, params, headers)
+        params['start-index'] = start_index
+        response = self.get(path, data, flags, params, headers)
+        if self.advanced_mode:
+            try:
+                response.raise_for_status()
+                response = response.json()
+            except HTTPError as e:
+                logging.error("Broken response: {}".format(e))
+                yield e
+        try:
             results = response[elements_key]
-            size = results['size']
-            # Check if start index was reset when reaching the end of the pages list
-            if results['start-index'] < start_index:
-                break
+            size = 0 
+            # Check if we still can get results
+            if size > max_results or results['size'] == 0:
+                return
             for r in results[element_key]:
+                size += 1
                 yield r
             start_index += results['max-result']
+        except TypeError:
+            logging.error("Broken response: {}".format(response))
+            yield response
+        
 
-    def base_list_call(self, resource, expand, favourite, clover_enabled, max_results, label=None, start_index=0,
-                       **kwargs):
+    def base_list_call(self, resource, expand, favourite, clover_enabled, max_results, label=None, start_index=0, **kwargs):
         flags = []
         params = {'max-results': max_results}
         if expand:
@@ -57,20 +73,12 @@ class Bamboo(AtlassianRestAPI):
         if 'elements_key' in kwargs and 'element_key' in kwargs:
             return self._get_generator(self.resource_url(resource), flags=flags, params=params,
                                        elements_key=kwargs['elements_key'],
-                                       element_key=kwargs['element_key'])
+                                       element_key=kwargs['element_key'],
+                                       max_results=max_results)
         params['start-index'] = start_index
         return self.get(self.resource_url(resource), flags=flags, params=params)
 
-    def plan_directory_info(self, plan_key):
-        """
-        Returns information about the directories where artifacts, build logs, and build results will be stored. 
-        Disabled by default. 
-        See https://confluence.atlassian.com/display/BAMBOO/Plan+directory+information+REST+API for more information.
-        :param plan_key:
-        :return:
-        """
-        resource = 'planDirectoryInfo/{}'.format(plan_key)
-        return self.get(self.resource_url(resource))
+    """ Projects & Plans """
 
     def projects(self, expand=None, favourite=False, clover_enabled=False, max_results=25):
         return self.base_list_call('project', expand, favourite, clover_enabled, max_results,
@@ -94,8 +102,112 @@ class Bamboo(AtlassianRestAPI):
         return self.base_list_call("plan", expand, favourite, clover_enabled, start_index, max_results,
                                    elements_key='plans', element_key='plan')
 
+    def plan_directory_info(self, plan_key):
+        """
+        Returns information about the directories where artifacts, build logs, and build results will be stored. 
+        Disabled by default. 
+        See https://confluence.atlassian.com/display/BAMBOO/Plan+directory+information+REST+API for more information.
+        :param plan_key:
+        :return:
+        """
+        resource = 'planDirectoryInfo/{}'.format(plan_key)
+        return self.get(self.resource_url(resource))
+
+    def get_plan(self, plan_key):
+        """
+        Get plan information.
+        :param plan_key:
+        :return:
+        """
+        resource = 'rest/api/latest/plan/{}'.format(plan_key)
+        return self.get(resource)
+
+    def delete_plan(self, plan_key):
+        """
+        Marks plan for deletion. Plan will be deleted by a batch job.
+        :param plan_key:
+        :return:
+        """
+        resource = 'rest/api/latest/plan/{}'.format(plan_key)
+        return self.delete(resource)
+
+    def enable_plan(self, plan_key):
+        """
+        Enable plan.
+        :param plan_key: str TST-BLD
+        :return: POST request
+        """
+        resource = 'plan/{plan_key}/enable'.format(plan_key=plan_key)
+        return self.post(self.resource_url(resource))
+
+    """ Branches """
+
+    def search_branches(self, plan_key, include_default_branch=True, max_results=25):
+        params = {
+            'max-result': max_results,
+            'start-index': 0,
+            'masterPlanKey': plan_key,
+            'includeMasterBranch': include_default_branch
+        }
+        size = 1
+        while params['start-index'] < size:
+            results = self.get(self.resource_url('search/branches'), params=params)
+            size = results['size']
+            for r in results['searchResults']:
+                yield r
+            params['start-index'] += results['max-result']
+
+    def plan_branches(self, plan_key, expand=None, favourite=False, clover_enabled=False, max_results=25):
+        """api/1.0/plan/{projectKey}-{buildKey}/branch"""
+        resource = 'plan/{}/branch'.format(plan_key)
+        return self.base_list_call(resource, expand, favourite, clover_enabled, max_results,
+                                   elements_key='branches', element_key='branch')
+
+    def get_branch_info(self, plan_key, branch_name):
+        """
+        Get information about a plan branch
+        :param plan_key: 
+        :param branch_name: 
+        :return:
+        """
+        resource = 'plan/{plan_key}/branch/{branch_name}'.format(plan_key=plan_key, branch_name=branch_name)
+        return self.get(self.resource_url(resource))
+
+    def create_branch(self, plan_key, branch_name, vcs_branch=None, enabled=False, cleanup_enabled=False):
+        """
+        Method for creating branch for a specified plan. 
+        You can use vcsBranch query param to define which vcsBranch should newly created branch use. 
+        If not specified it will not override vcsBranch from the main plan. 
+
+        :param plan_key: str TST-BLD
+        :param branch_name: str new-shiny-branch
+        :param vcs_branch: str feature/new-shiny-branch, /refs/heads/new-shiny-branch
+        :param enabled: bool
+        :param cleanup_enabled: bool
+        :return: PUT request
+        """
+        resource = 'plan/{plan_key}/branch/{branch_name}'.format(plan_key=plan_key, branch_name=branch_name)
+        params = {}
+        if vcs_branch:
+            params = dict(vcsBranch=vcs_branch,
+                          enabled='true' if enabled else 'false',
+                          cleanupEnabled='true' if cleanup_enabled else 'false')
+        return self.put(self.resource_url(resource), params=params)
+
+    def get_vcs_branches(self, plan_key, max_results=25):
+        """
+        Get all vcs names for the current plan
+        :param plan_key: str TST-BLD
+        :param max_results
+        :return:
+        """
+        resource = 'plan/{plan_key}/vcsBranches'.format(plan_key=plan_key)
+        return self.base_list_call(resource, start_index=0, max_results=max_results)
+
+    """ Build results """
+
     def results(self, project_key=None, plan_key=None, job_key=None, build_number=None, expand=None, favourite=False,
-                clover_enabled=False, issue_key=None, label=None, start_index=0, max_results=25):
+                clover_enabled=False, issue_key=None, label=None, start_index=0, max_results=25, include_all_states=False):
         """
         Get results as generic method
         :param project_key:
@@ -109,6 +221,7 @@ class Bamboo(AtlassianRestAPI):
         :param label:
         :param start_index:
         :param max_results:
+        :param include_all_states:
         :return:
         """
         resource = "result"
@@ -124,12 +237,14 @@ class Bamboo(AtlassianRestAPI):
         params = {}
         if issue_key:
             params['issueKey'] = issue_key
+        if include_all_states:
+            params['includeAllStates'] = include_all_states
         return self.base_list_call(resource, expand=expand, favourite=favourite, clover_enabled=clover_enabled,
                                    start_index=start_index, max_results=max_results,
                                    elements_key='results', element_key='result', label=label, **params)
 
     def latest_results(self, expand=None, favourite=False, clover_enabled=False, label=None, issue_key=None,
-                       start_index=0, max_results=25):
+                       start_index=0, max_results=25, include_all_states=False):
         """
         Get latest Results
         :param expand:
@@ -139,13 +254,14 @@ class Bamboo(AtlassianRestAPI):
         :param issue_key:
         :param start_index:
         :param max_results:
+        :param include_all_states:
         :return:
         """
         return self.results(expand=expand, favourite=favourite, clover_enabled=clover_enabled,
-                            label=label, issue_key=issue_key, start_index=start_index, max_results=max_results)
+                            label=label, issue_key=issue_key, start_index=start_index, max_results=max_results, include_all_states=include_all_states)
 
     def project_latest_results(self, project_key, expand=None, favourite=False, clover_enabled=False, label=None,
-                               issue_key=None, start_index=0, max_results=25):
+                               issue_key=None, start_index=0, max_results=25, include_all_states=False):
         """
         Get latest Project Results
         :param project_key:
@@ -156,13 +272,14 @@ class Bamboo(AtlassianRestAPI):
         :param issue_key:
         :param start_index:
         :param max_results:
+        :param include_all_states:
         :return:
         """
         return self.results(project_key, expand=expand, favourite=favourite, clover_enabled=clover_enabled,
-                            label=label, issue_key=issue_key, start_index=start_index, max_results=max_results)
+                            label=label, issue_key=issue_key, start_index=start_index, max_results=max_results, include_all_states=include_all_states)
 
     def plan_results(self, project_key, plan_key, expand=None, favourite=False, clover_enabled=False, label=None,
-                     issue_key=None, start_index=0, max_results=25):
+                     issue_key=None, start_index=0, max_results=25, include_all_states=False):
         """
         Get Plan results
         :param project_key:
@@ -174,12 +291,13 @@ class Bamboo(AtlassianRestAPI):
         :param issue_key:
         :param start_index:
         :param max_results:
+        :param include_all_states:
         :return:
         """
         return self.results(project_key, plan_key, expand=expand, favourite=favourite, clover_enabled=clover_enabled,
-                            label=label, issue_key=issue_key, start_index=start_index, max_results=max_results)
+                            label=label, issue_key=issue_key, start_index=start_index, max_results=max_results, include_all_states=include_all_states)
 
-    def build_result(self, build_key, expand=None):
+    def build_result(self, build_key, expand=None, include_all_states=False):
         """
         Returns details of a specific build result
         :param expand: expands build result details on request. Possible values are: artifacts, comments, labels,
@@ -187,27 +305,29 @@ class Bamboo(AtlassianRestAPI):
         using stages.stage.results.result. All expand parameters should contain results.result prefix.
         :param build_key: Should be in the form XX-YY[-ZZ]-99, that is, the last token should be an integer representing
         the build number
+        :param include_all_states
         """
         try:
             int(build_key.split('-')[-1])
             resource = "result/{}".format(build_key)
             return self.base_list_call(resource, expand, favourite=False, clover_enabled=False,
-                                       start_index=0, max_results=25)
+                                       start_index=0, max_results=25, include_all_states=include_all_states)
         except ValueError:
             raise ValueError('The key "{}" does not correspond to a build result'.format(build_key))
 
-    def build_latest_result(self, plan_key, expand=None):
+    def build_latest_result(self, plan_key, expand=None, include_all_states=False):
         """
         Returns details of a latest build result
         :param expand: expands build result details on request. Possible values are: artifacts, comments, labels,
         Jira Issues, stages. stages expand is available only for top level plans. It allows to drill down to job results
         using stages.stage.results.result. All expand parameters should contain results.result prefix.
         :param plan_key: Should be in the form XX-YY[-ZZ]
+        :param include_all_states:
         """
         try:
             resource = "result/{}/latest.json".format(plan_key)
             return self.base_list_call(resource, expand, favourite=False, clover_enabled=False,
-                                       start_index=0, max_results=25)
+                                       start_index=0, max_results=25, include_all_states=include_all_states)
         except ValueError:
             raise ValueError('The key "{}" does not correspond to the latest build result'.format(plan_key))
 
@@ -223,34 +343,33 @@ class Bamboo(AtlassianRestAPI):
         params = {'buildKey': plan_key, 'buildNumber': build_number}
         return self.post(custom_resource, params=params, headers=self.form_token_headers)
 
-    def delete_plan(self, plan_key):
+    def execute_build(self, plan_key, stage=None, execute_all_stages=True, custom_revision=None, **bamboo_variables):
         """
-        Marks plan for deletion. Plan will be deleted by a batch job.
-        :param plan_key:
-        :return:
+        Fire build execution for specified plan. 
+        !IMPORTANT! NOTE: for some reason, this method always execute all stages
+        :param plan_key: str TST-BLD
+        :param stage: str stage-name
+        :param execute_all_stages: bool
+        :param custom_revision: str revisionName
+        :param bamboo_variables: dict {variable=value} 
+        :return: POST request
         """
-        resource = 'rest/api/latest/plan/{}'.format(plan_key)
-        return self.delete(resource)
+        headers = self.form_token_headers
+        resource = 'queue/{plan_key}'.format(plan_key=plan_key)
+        params = {}
+        if stage:
+            execute_all_stages = False
+            params['stage'] = stage
+        if custom_revision:
+            params['customRevision'] = custom_revision
+        params['executeAllStages'] = 'true' if execute_all_stages else 'false'
+        if bamboo_variables:
+            for key, value in bamboo_variables.items():
+                params['bamboo.variable.{}'.format(key)] = value
 
-    def reports(self, max_results=25):
-        params = {'max-results': max_results}
-        return self._get_generator(self.resource_url('chart/reports'), elements_key='reports', element_key='report',
-                                   params=params)
+        return self.post(self.resource_url(resource), params=params, headers=headers)
 
-    def chart(self, report_key, build_keys, group_by_period, date_filter=None, date_from=None, date_to=None,
-              width=None, height=None, start_index=9, max_results=25):
-        params = {'reportKey': report_key, 'buildKeys': build_keys, 'groupByPeriod': group_by_period,
-                  'start-index': start_index, 'max-results': max_results}
-        if date_filter:
-            params['dateFilter'] = date_filter
-            if date_filter == 'RANGE':
-                params['dateFrom'] = date_from
-                params['dateTo'] = date_to
-        if width:
-            params['width'] = width
-        if height:
-            params['height'] = height
-        return self.get(self.resource_url('chart'), params=params)
+    """ Comments & Labels """
 
     def comments(self, project_key, plan_key, build_number, start_index=0, max_results=25):
         resource = "result/{}-{}-{}/comment".format(project_key, plan_key, build_number)
@@ -275,24 +394,17 @@ class Bamboo(AtlassianRestAPI):
         resource = "result/{}-{}-{}/label/{}".format(project_key, plan_key, build_number, label)
         return self.delete(self.resource_url(resource))
 
-    def server_info(self):
-        return self.get(self.resource_url('info'))
-
-    def agent_status(self):
-        return self.get(self.resource_url('agent'))
-
-    def activity(self):
-        return self.get('build/admin/ajax/getDashboardSummary.action')
-
-    def deployment_project(self, project_id):
-        resource = 'deploy/project/{}'.format(project_id)
-        return self.get(self.resource_url(resource))
+    """ Deployments """
 
     def deployment_projects(self):
         resource = 'deploy/project/all'
         for project in self.get(self.resource_url(resource)):
             yield project
 
+    def deployment_project(self, project_id):
+        resource = 'deploy/project/{}'.format(project_id)
+        return self.get(self.resource_url(resource))
+
     def deployment_environment_results(self, env_id, expand=None, max_results=25):
         resource = 'deploy/environment/{environmentId}/results'.format(environmentId=env_id)
         params = {'max-result': max_results, 'start-index': 0}
@@ -314,39 +426,7 @@ class Bamboo(AtlassianRestAPI):
         resource = 'deploy/dashboard/{}'.format(project_id) if project_id else 'deploy/dashboard'
         return self.get(self.resource_url(resource))
 
-    def search_branches(self, plan_key, include_default_branch=True, max_results=25):
-        params = {
-            'max-result': max_results,
-            'start-index': 0,
-            'masterPlanKey': plan_key,
-            'includeMasterBranch': include_default_branch
-        }
-        size = 1
-        while params['start-index'] < size:
-            results = self.get(self.resource_url('search/branches'), params=params)
-            size = results['size']
-            for r in results['searchResults']:
-                yield r
-            params['start-index'] += results['max-result']
-
-    def plan_branches(self, plan_key, expand=None, favourite=False, clover_enabled=False, max_results=25):
-        """api/1.0/plan/{projectKey}-{buildKey}/branch"""
-        resource = 'plan/{}/branch'.format(plan_key)
-        return self.base_list_call(resource, expand, favourite, clover_enabled, max_results,
-                                   elements_key='branches', element_key='branch')
-
-    def health_check(self):
-        """
-        Get health status
-        https://confluence.atlassian.com/jirakb/how-to-retrieve-health-check-results-using-rest-api-867195158.html
-        :return:
-        """
-        # check as Troubleshooting & Support Tools Plugin
-        response = self.get('rest/troubleshooting/1.0/check/')
-        if not response:
-            # check as support tools
-            response = self.get('rest/supportHealthCheck/1.0/check/')
-        return response
+    """ Users & Groups """
 
     def get_users_in_global_permissions(self, start=0, limit=25):
         """
@@ -404,7 +484,7 @@ class Bamboo(AtlassianRestAPI):
         url = 'rest/api/latest/admin/groups/{}/add-users'.format(group_name)
         return self.post(url, data=users)
 
-    def remove_users_into_group(self, group_name, users):
+    def remove_users_from_group(self, group_name, users):
         """
         Remove multiple users from a group.
         The list of usernames should be passed as request body.
@@ -457,3 +537,73 @@ class Bamboo(AtlassianRestAPI):
         """
         params = {'expand': expand}
         return self.get('rest/api/latest/queue', params=params)
+
+    """Other actions"""
+
+    def server_info(self):
+        return self.get(self.resource_url('info'))
+
+    def agent_status(self):
+        return self.get(self.resource_url('agent'))
+
+    def activity(self):
+        return self.get('build/admin/ajax/getDashboardSummary.action')
+
+    def get_custom_expiry(self, limit=25):
+        """
+        Get list of all plans where user has admin permission and which override global expiry settings. 
+        If global expiry is not enabled it returns empty response.
+        :param limit:
+        """
+        url = "rest/api/latest/admin/expiry/custom/plan?limit={}".format(limit)
+        return self.get(url)
+
+    def reports(self, max_results=25):
+        params = {'max-results': max_results}
+        return self._get_generator(self.resource_url('chart/reports'), elements_key='reports', element_key='report',
+                                   params=params)
+
+    def chart(self, report_key, build_keys, group_by_period, date_filter=None, date_from=None, date_to=None,
+              width=None, height=None, start_index=9, max_results=25):
+        params = {'reportKey': report_key, 'buildKeys': build_keys, 'groupByPeriod': group_by_period,
+                  'start-index': start_index, 'max-results': max_results}
+        if date_filter:
+            params['dateFilter'] = date_filter
+            if date_filter == 'RANGE':
+                params['dateFrom'] = date_from
+                params['dateTo'] = date_to
+        if width:
+            params['width'] = width
+        if height:
+            params['height'] = height
+        return self.get(self.resource_url('chart'), params=params)
+
+    def health_check(self):
+        """
+        Get health status
+        https://confluence.atlassian.com/jirakb/how-to-retrieve-health-check-results-using-rest-api-867195158.html
+        :return:
+        """
+        # check as Troubleshooting & Support Tools Plugin
+        response = self.get('rest/troubleshooting/1.0/check/')
+        if not response:
+            # check as support tools
+            response = self.get('rest/supportHealthCheck/1.0/check/')
+        return response
+
+    def upload_plugin(self, plugin_path):
+        """
+        Provide plugin path for upload into Jira e.g. useful for auto deploy
+        :param plugin_path:
+        :return:
+        """
+        files = {
+            'plugin': open(plugin_path, 'rb')
+        }
+        headers = {
+            'X-Atlassian-Token': 'nocheck'
+        }
+        upm_token = self.request(method='GET', path='rest/plugins/1.0/', headers=headers, trailing=True).headers[
+            'upm-token']
+        url = 'rest/plugins/1.0/?token={upm_token}'.format(upm_token=upm_token)
+        return self.post(url, files=files, headers=headers)
