@@ -1,11 +1,13 @@
 # coding=utf-8
 import logging
+
 from .rest_client import AtlassianRestAPI
 
 log = logging.getLogger(__name__)
 
 
 class Bitbucket(AtlassianRestAPI):
+
     def project_list(self, limit=None):
         """
         Provide the project list
@@ -17,6 +19,16 @@ class Bitbucket(AtlassianRestAPI):
             params['limit'] = limit
         return (self.get('rest/api/1.0/projects', params=params) or {}).get('values')
 
+    def repo_list(self, project_key):
+        """
+        Provide the repository list in a specific BB project
+        :param project_key: The Project Key ID you need to list
+        return:
+        """
+        url = 'rest/api/1.0/projects/{projectKey}/repos?limit=1000'.format(
+            projectKey=project_key)
+        return self.get(url)
+
     def project(self, key):
         """
         Provide project info
@@ -24,7 +36,7 @@ class Bitbucket(AtlassianRestAPI):
         :return:
         """
         url = 'rest/api/1.0/projects/{0}'.format(key)
-        return (self.get(url) or {}).get('values')
+        return self.get(url) or {}
 
     def create_project(self, key, name, description=""):
         """
@@ -41,6 +53,48 @@ class Bitbucket(AtlassianRestAPI):
                 }
         return self.post(url, data=data)
 
+    def update_project(self, key, **params):
+        """
+        Update project
+        :param key:
+        :param **params:
+        :return:
+        """
+        data = self.project(key)
+        if not 'errors' in data:
+            data.update(params)
+            url = 'rest/api/1.0/projects/{0}'.format(key)
+            return self.put(url, data=data)
+        else:
+            log.debug('Failed to update project: {0}: Unable to read project'.format(key))
+            return None
+
+    def project_avatar(self, key, content_type='image/png'):
+        """
+        Get project avatar
+
+        :param key:
+        :return:
+        """
+        url = 'rest/api/1.0/projects/{0}/avatar.png'.format(key)
+        headers = dict(self.default_headers)
+        headers['Accept'] = content_type
+        headers['X-Atlassian-Token'] = 'no-check'
+
+        return self.get(url, not_json_response=True, headers=headers) or {}
+
+    def set_project_avatar(self, key, icon, content_type='image/png'):
+        """
+        Set project avatar
+
+        :param key:
+        :return:
+        """
+        headers = {'X-Atlassian-Token': 'no-check'}
+        files = {'avatar': ("avatar.png", icon, content_type)}
+        url = 'rest/api/1.0/projects/{0}/avatar.png'.format(key)
+        return self.post(url, files=files, headers=headers) or {}
+
     def project_users(self, key, limit=99999, filter_str=None):
         """
         Get users who has permission in project
@@ -97,7 +151,7 @@ class Bitbucket(AtlassianRestAPI):
         :param project_key: project key involved
         :param username: user name to be granted
         :param permission: the project permissions available are 'PROJECT_ADMIN', 'PROJECT_WRITE' and 'PROJECT_READ'
-        :return: 
+        :return:
         """
         url = 'rest/api/1.0/projects/{project_key}/permissions/users?permission={permission}&name={username}'.format(
             project_key=project_key,
@@ -126,7 +180,7 @@ class Bitbucket(AtlassianRestAPI):
         :param project_key: project key involved
         :param groupname: group to be granted
         :param permission: the project permissions available are 'PROJECT_ADMIN', 'PROJECT_WRITE' and 'PROJECT_READ'
-        :return: 
+        :return:
         """
         url = 'rest/api/1.0/projects/{project_key}/permissions/groups?permission={permission}&name={groupname}'.format(
             project_key=project_key,
@@ -157,7 +211,7 @@ class Bitbucket(AtlassianRestAPI):
         :param repo_key: repository key involved (slug)
         :param username: user name to be granted
         :param permission: the repository permissions available are 'REPO_ADMIN', 'REPO_WRITE' and 'REPO_READ'
-        :return: 
+        :return:
         """
         params = {'permission': permission,
                   'name': username}
@@ -198,7 +252,7 @@ class Bitbucket(AtlassianRestAPI):
         :param repo_key: repository key involved (slug)
         :param groupname: group to be granted
         :param permission: the repository permissions available are 'REPO_ADMIN', 'REPO_WRITE' and 'REPO_READ'
-        :return: 
+        :return:
         """
         params = {'permission': permission,
                   'name': groupname}
@@ -273,7 +327,8 @@ class Bitbucket(AtlassianRestAPI):
             'key': key,
             'data': self.project(key),
             'users': self.project_users(key),
-            'groups': self.project_groups(key)}
+            'groups': self.project_groups(key),
+            'avatar': self.project_avatar(key)}
 
     def group_members(self, group, limit=99999):
         """
@@ -359,6 +414,19 @@ class Bitbucket(AtlassianRestAPI):
         }
         return self.post(url, data=data)
 
+    def get_repo(self, project_key, repository_slug):
+        """
+        Get a specific repository from a project. This operates based on slug not name which may
+        be confusing to some users.
+        :param project_key: Key of the project you wish to look in.
+        :param repository_slug: url-compatible repository identifier
+        :return: Dictionary of request response
+        """
+
+        url = 'rest/api/1.0/projects/{project}/repos/{repository}'.format(project=project_key,
+                                                                          repository=repository_slug)
+        return self.get(url)
+
     def repo_all_list(self, project_key):
         """
         Get all repositories list from project
@@ -380,6 +448,19 @@ class Bitbucket(AtlassianRestAPI):
             repo_list += (response or {}).get('values')
         return repo_list
 
+    def delete_repo(self, project_key, repository_slug):
+        """
+        Delete a specific repository from a project. This operates based on slug not name which may
+        be confusing to some users.
+        :param project_key: Key of the project you wish to look in.
+        :param repository_slug: url-compatible repository identifier
+        :return: Dictionary of request response
+        """
+
+        url = 'rest/api/1.0/projects/{project}/repos/{repository}'.format(project=project_key,
+                                                                          repository=repository_slug)
+        return self.delete(url)
+
     def get_branches(self, project, repository, base=None, filter=None, start=0, limit=99999, details=True,
                      order_by='MODIFICATION'):
         """
@@ -473,7 +554,30 @@ class Bitbucket(AtlassianRestAPI):
         data = {"name": str(name), "endPoint": str(end_point)}
         return self.delete(url, data=data)
 
-    def get_pull_requests(self, project, repository, state='OPEN', order='newest', limit=100, start=0):
+    def get_pull_request_settings(self, project, repository):
+        """
+        Get pull request settings.
+        :param project:
+        :param repository:
+        :return:
+        """
+        url = 'rest/api/1.0/projects/{project}/repos/{repository}/settings/pull-requests'.format(project=project,
+                                                                                                 repository=repository)
+        return self.get(url)
+
+    def set_pull_request_settings(self, project, repository, data):
+        """
+        Set pull request settings.
+        :param project:
+        :param repository:
+        :param data: json body
+        :return:
+        """
+        url = 'rest/api/1.0/projects/{project}/repos/{repository}/settings/pull-requests'.format(project=project,
+                                                                                                 repository=repository)
+        return self.post(url, data=data)
+
+    def get_pull_requests(self, project, repository, state='OPEN', order='newest', limit=100, start=0, at=None):
         """
         Get pull requests
         :param project:
@@ -496,6 +600,8 @@ class Bitbucket(AtlassianRestAPI):
             params['start'] = start
         if order:
             params['order'] = order
+        if at:
+            params['at'] = at
         response = self.get(url, params=params)
         if 'values' not in response:
             return []
@@ -653,13 +759,15 @@ class Bitbucket(AtlassianRestAPI):
         url = 'rest/api/1.0/inbox/pull-requests'
         return self.get(url, params=params)
 
-    def add_pull_request_comment(self, project, repository, pull_request_id, text):
+    def add_pull_request_comment(self, project, repository, pull_request_id, text, parent_id=None):
         """
         Add comment into pull request
         :param project:
         :param repository:
         :param pull_request_id: the ID of the pull request within the repository
         :param text comment text
+        :param parent_id parent comment id
+
         :return:
         """
         url = 'rest/api/1.0/projects/{project}/repos/{repository}/pull-requests/{pullRequestId}/comments'.format(
@@ -667,8 +775,47 @@ class Bitbucket(AtlassianRestAPI):
             repository=repository,
             pullRequestId=pull_request_id)
         body = {'text': text}
+        if parent_id:
+            body['parent'] = {'id': parent_id}
         return self.post(url, data=body)
 
+    def get_pull_request_comment(self, project, repository, pull_request_id, comment_id):
+        """
+        Retrieves a pull request comment.
+        The authenticated user must have REPO_READ permission
+        for the repository that this pull request targets to call this resource.
+        :param project:
+        :param repository:
+        :param pull_request_id: the ID of the pull request within the repository
+        :param comment_id: the ID of the comment to retrieve
+        :return:
+        """
+        url = ('rest/api/1.0/projects/{project}/repos/{repository}'
+               '/pull-requests/{pullRequestId}/comments/{comment_id}').format(
+            project=project,
+            repository=repository,
+            pullRequestId=pull_request_id,
+            comment_id=comment_id)
+        return self.get(url)
+
+    def update_pull_request_comment(self, project, repository, pull_request_id, comment_id, comment, comment_version):
+        """
+        Update the text of a comment.
+        Only the user who created a comment may update it.
+
+        Note: the supplied supplied JSON object must contain a version
+        that must match the server's version of the comment
+        or the update will fail.
+        """
+        url = "/rest/api/1.0/projects/{project}/repos/{repository}/pull-requests/{pull_request_id}/comments/{comment_id}".format(
+            project=project, repository=repository, pull_request_id=pull_request_id, comment_id=comment_id
+        )
+        payload = {
+            "version": comment_version,
+            "text": comment
+        }
+        return self.put(url, data=payload)
+
     def get_pullrequest(self, project, repository, pull_request_id):
         """
         Retrieve a pull request.
@@ -685,6 +832,31 @@ class Bitbucket(AtlassianRestAPI):
             pullRequestId=pull_request_id)
         return self.get(url)
 
+    def change_reviewed_status(self, project_key, repository_slug, pull_request_id, status, user_slug):
+        """
+        Change the current user's status for a pull request.
+        Implicitly adds the user as a participant if they are not already.
+        If the current user is the author, this method will fail.
+        :param project_key:
+        :param repository_slug:
+        :param pull_request_id:
+        :param status:
+        :param user_slug:
+        :return:
+        """
+        url = "rest/api/1.0/projects/{projectKey}/repos/{repositorySlug}/pull-requests/{pullRequestId}/participants/{userSlug}".format(
+            projectKey=project_key, repositorySlug=repository_slug, pullRequestId=pull_request_id, userSlug=user_slug,
+        )
+        approved = True if status == "APPROVED" else False
+        data = {
+            "user": {
+                "name": user_slug
+            },
+            "approved": approved,
+            "status": status
+        }
+        return self.put(url, data)
+
     def get_tags(self, project, repository, filter='', limit=1000, order_by=None, start=0):
         """
         Retrieve the tags matching the supplied filterText param.
@@ -896,22 +1068,47 @@ class Bitbucket(AtlassianRestAPI):
             params['markup'] = markup
         return self.get(url, params=params, not_json_response=True, headers=headers)
 
-    def get_branches_permissions(self, project, repository, limit=25):
+    def get_branches_permissions(self, project, repository=None, start=0, limit=25):
         """
         Get branches permissions from a given repo
         :param project:
         :param repository:
+        :param start:
         :param limit:
         :return:
         """
-        url = 'rest/branch-permissions/2.0/projects/{project}/repos/{repository}/restrictions'.format(
-            project=project,
-            repository=repository)
+        if repository is not None:
+            url = 'rest/branch-permissions/2.0/projects/{project}/repos/{repository}/restrictions'.format(
+                project=project,
+                repository=repository)
+        else:
+            url = 'rest/branch-permissions/2.0/projects/{project}/restrictions'.format(
+                project=project)
+
         params = {}
         if limit:
             params['limit'] = limit
+        if start:
+            params['start'] = start
         return self.get(url, params=params)
 
+    def all_branches_permissions(self, project, repository=None):
+        """
+        Get branches permissions from a given repo
+        :param project:
+        :param repository:
+        :return:
+        """
+        start = 0
+        branches_permissions = []
+        response = self.get_branches_permissions(project=project, repository=repository, start=start)
+        branches_permissions += response.get('values')
+        while not response.get('isLastPage'):
+            start = response.get('nextPageStart')
+            response = self.get_branches_permissions(project=project, repository=repository, start=start)
+            branches_permissions += response.get('values')
+        return branches_permissions
+
     def reindex(self):
         """
         Rebuild the bundled Elasticsearch indexes for Bitbucket Server
@@ -1103,3 +1300,111 @@ class Bitbucket(AtlassianRestAPI):
         :return:
         """
         return self.post('rest/api/1.0/markup/preview', data=data)
+
+    def upload_plugin(self, plugin_path):
+        """
+        Provide plugin path for upload into Jira e.g. useful for auto deploy
+        :param plugin_path:
+        :return:
+        """
+        files = {
+            'plugin': open(plugin_path, 'rb')
+        }
+        headers = {
+            'X-Atlassian-Token': 'nocheck'
+        }
+        upm_token = self.request(method='GET', path='rest/plugins/1.0/', headers=headers, trailing=True).headers[
+            'upm-token']
+        url = 'rest/plugins/1.0/?token={upm_token}'.format(upm_token=upm_token)
+        return self.post(url, files=files, headers=headers)
+
+    def upload_file(self, project, repository, content, message, branch, filename):
+        """
+        Upload new file for given branch.
+        :param content:
+        :param message:
+        :param branch:
+        :return:
+        """
+        data = {
+            "content": content,
+            "message": message,
+            "branch": branch
+        }
+
+        url = 'rest/api/1.0/projects/{project}/repos/{repository}/browse/{filename}'.format(
+            project=project,
+            repository=repository,
+            filename=filename)
+        return self.put(url, files=data)
+
+    def update_file(self, project, repository, content, message, branch, filename, source_commit_id):
+        """
+        Update existing file for given branch.
+        :param project:
+        :param repository:
+        :param content:
+        :param message:
+        :param branch:
+        :param filename:
+        :param source_commit_id:
+        :return:
+        """
+        data = {
+            "content": content,
+            "message": message,
+            "branch": branch,
+            "sourceCommitId": source_commit_id
+        }
+
+        url = 'rest/api/1.0/projects/{project}/repos/{repository}/browse/{filename}'.format(
+            project=project,
+            repository=repository,
+            filename=filename)
+        return self.put(url, files=data)
+
+    def get_code_insights_report(self, project_key, repository_slug, commit_id, report_key):
+        """
+        Retrieve the specified code-insights report.
+        :projectKey: str
+        :repositorySlug: str
+        :commitId: str
+        :report_key: str
+        """
+        url = "rest/insights/1.0/projects/{projectKey}/repos/{repositorySlug}/commits/{commitId}/reports/{key}".format(
+            projectKey=project_key, repositorySlug=repository_slug, commitId=commit_id, key=report_key
+        )
+        return self.get(url)
+
+    def delete_code_insights_report(self, project_key, repository_slug, commit_id, report_key):
+        """
+        Delete a report for the given commit. Also deletes any annotations associated with this report.
+        :projectKey: str
+        :repositorySlug: str
+        :commitId: str
+        :report_key: str
+        """
+        url = "rest/insights/1.0/projects/{projectKey}/repos/{repositorySlug}/commits/{commitId}/reports/{key}".format(
+            projectKey=project_key, repositorySlug=repository_slug, commitId=commit_id, key=report_key
+        )
+        return self.delete(url)
+
+    def create_code_insights_report(self, project_key, repository_slug, commit_id, report_key, report_title,
+                                    **report_params):
+        """
+        Create a new insight report, or replace the existing one if a report already exists for the given repository, commit, and report key. 
+        A request to replace an existing report will be rejected if the authenticated user was not the creator of the specified report.
+        For further information visit: https://docs.atlassian.com/bitbucket-server/rest/6.6.1/bitbucket-code-insights-rest.html
+        :projectKey: str
+        :repositorySlug: str
+        :commitId: str
+        :report_key: str
+        :report_title: str
+        :report_params: 
+        """
+        url = "rest/insights/1.0/projects/{projectKey}/repos/{repositorySlug}/commits/{commitId}/reports/{key}".format(
+            projectKey=project_key, repositorySlug=repository_slug, commitId=commit_id, key=report_key
+        )
+        data = {"title": report_title}
+        data.update(report_params)
+        return self.put(url, data=data)
