@@ -1,6 +1,8 @@
 # coding=utf-8
 import logging
+
 from requests.exceptions import HTTPError
+
 from .rest_client import AtlassianRestAPI
 
 log = logging.getLogger(__name__)
@@ -205,6 +207,19 @@ class Jira(AtlassianRestAPI):
         params = {'username': username}
         return self.delete(url, params=params)
 
+    def user_update_or_create_property_through_rest_point(self, username, key, value):
+        """
+        ATTENTION!
+        This method used after configuration of rest endpoint on Jira side
+        :param username:
+        :param key:
+        :param value:
+        :return:
+        """
+        url = 'rest/scriptrunner/latest/custom/updateUserProperty'
+        params = {'username': username, 'property': key, 'value': value}
+        return self.get(url, params=params)
+
     def user_update_email(self, username, email):
         """
         Update user email for new domain changes
@@ -363,6 +378,20 @@ class Jira(AtlassianRestAPI):
             params['expand'] = expand
         return self.get('rest/api/2/project/{}/version'.format(key), params)
 
+    def add_version(self, project_key, project_id, version, is_archived=False, is_released=False):
+        """
+        Add missing version to project
+        :param project_key: the project key
+        :param project_id: the project id
+        :param version: the new project version to add
+        :is_archived:
+        :is_released:
+        :return:
+        """
+        payload = {'name': version, 'archived': is_archived, 'released': is_released, 'project': project_key,
+                   'projectId': project_id}
+        return self.post("rest/api/2/version", data=payload)
+
     def get_project_roles(self, project_key):
         """
         Provide associated project roles
@@ -550,6 +579,22 @@ class Jira(AtlassianRestAPI):
         if limit:
             params['maxResults'] = limit
         return self.get(url, params=params)
+    
+    def create_custom_field(self, name, type, search_key=None, description=None):
+        """
+        Creates a custom field with the given name and type
+        :param name: str
+        :param type: str, like 'com.atlassian.jira.plugin.system.customfieldtypes:textfield'
+        :param search_key: str, like above
+        :param description: str
+        """
+        url = 'rest/api/2/field'
+        data = {'name': name, 'type': type}
+        if search_key:
+            data['search_key'] = search_key
+        if description:
+            data['description'] = description
+        return self.post(url, data=data)
 
     def get_all_available_screen_fields(self, screen_id):
         """
@@ -963,9 +1008,28 @@ class Jira(AtlassianRestAPI):
         return self.delete(url)
 
     def get_issue_transitions(self, issue_key):
-        url = 'rest/api/2/issue/{issue_key}?expand=transitions.fields&fields=status'.format(issue_key=issue_key)
         return [{'name': transition['name'], 'id': int(transition['id']), 'to': transition['to']['name']}
-                for transition in (self.get(url) or {}).get('transitions')]
+                for transition in (self.get_issue_transitions_full(issue_key) or {}).get('transitions')]
+
+    def get_issue_transitions_full(self, issue_key, transition_id=None, expand=None):
+        """
+        Get a list of the transitions possible for this issue by the current user,
+        along with fields that are required and their types.
+        Fields will only be returned if expand = 'transitions.fields'.
+        The fields in the metadata correspond to the fields in the transition screen for that transition.
+        Fields not in the screen will not be in the metadata.
+        :param issue_key: str
+        :param transition_id: str
+        :param expand: str
+        :return:
+        """
+        url = 'rest/api/2/issue/{issue_key}/transitions'.format(issue_key=issue_key)
+        params = {}
+        if transition_id:
+            params['transitionId'] = transition_id
+        if expand:
+            params['expand'] = expand
+        return self.get(url, params=params)
 
     def get_status_id_from_name(self, status_name):
         url = 'rest/api/2/status/{name}'.format(name=status_name)
@@ -999,7 +1063,7 @@ class Jira(AtlassianRestAPI):
 
     def get_issue_status(self, issue_key):
         url = 'rest/api/2/issue/{issue_key}?fields=status'.format(issue_key=issue_key)
-        return (self.get(url) or {}).get('fields').get('status').get('name')
+        return (((self.get(url) or {}).get('fields') or {}).get('status') or {}).get('name') or {}
 
     def get_issue_status_id(self, issue_key):
         url = 'rest/api/2/issue/{issue_key}?fields=status'.format(issue_key=issue_key)
@@ -1611,6 +1675,17 @@ class Jira(AtlassianRestAPI):
         url = 'rest/tempo-core/1/workloadscheme/users/{}'.format(scheme_id)
         return self.get(url)
 
+    def tempo_workload_scheme_set_member(self, scheme_id, member):
+        """
+        Provide a workload scheme members
+        :param member: user name of user
+        :param scheme_id:
+        :return:
+        """
+        url = 'rest/tempo-core/1/workloadscheme/user/{}'.format(member)
+        data = {'id': scheme_id}
+        return self.put(url, data=data)
+
     def tempo_timesheets_get_configuration(self):
         """
         Provide the configs of timesheets
@@ -1666,12 +1741,12 @@ class Jira(AtlassianRestAPI):
 
     def tempo_timesheets_write_worklog(self, worker, started, time_spend_in_seconds, issue_id, comment=None):
         """
-
-        :param comment:
+        Log work for user
         :param worker:
         :param started:
         :param time_spend_in_seconds:
         :param issue_id:
+        :param comment:
         :return:
         """
         data = {"worker": worker,
@@ -1683,6 +1758,47 @@ class Jira(AtlassianRestAPI):
         url = 'rest/tempo-timesheets/4/worklogs/'
         return self.post(url, data=data)
 
+    def tempo_timesheets_approval_worklog_report(self, user_key, period_start_date):
+        """
+        Return timesheets for approval
+        :param user_key:
+        :param period_start_date:
+        :return:
+        """
+        url = "rest/tempo-timesheets/4/timesheet-approval/current"
+        params = {}
+        if period_start_date:
+            params['periodStartDate'] = period_start_date
+        if user_key:
+            params['userKey'] = user_key
+
+    def tempo_timesheets_get_required_times(self, from_date, to_date, user_name):
+        """
+        Provide time how much should work
+        :param from_date:
+        :param to_date:
+        :param user_name:
+        :return:
+        """
+        url = 'rest/tempo-timesheets/3/private/days'
+        params = {}
+        if from_date:
+            params['from'] = from_date
+        if to_date:
+            params['to'] = to_date
+        if user_name:
+            params['user'] = user_name
+        return self.get(url, params=params)
+
+    def tempo_timesheets_approval_status(self, period_start_date, user_name):
+        url = 'rest/tempo-timesheets/4/timesheet-approval/approval-statuses'
+        params = {}
+        if user_name:
+            params['userKey'] = user_name
+        if period_start_date:
+            params['periodStartDate'] = period_start_date
+        return self.get(url, params=params)
+
     def tempo_get_links_to_project(self, project_id):
         """
         Gets all links to a specific project
@@ -1701,6 +1817,13 @@ class Jira(AtlassianRestAPI):
         url = 'rest/tempo-accounts/1/link/project/{}/default/'.format(project_id)
         return self.get(url)
 
+    def tempo_teams_get_all_teams(self, expand=None):
+        url = "rest/tempo-teams/2/team"
+        params = {}
+        if expand:
+            params['expand'] = expand
+        return self.get(url, params=params)
+
     def tempo_teams_add_member(self, team_id, member_key):
         """
         Add team member
@@ -1708,12 +1831,37 @@ class Jira(AtlassianRestAPI):
         :param member_key:
         :return:
         """
-        url = 'rest/tempo-teams/2/team/{}/member/'.format(team_id)
-        data = {"member": {"key": member_key, "type": "USER"},
+        data = {"member": {"key": str(member_key), "type": "USER"},
                 "membership": {
                     "availability": "100",
                     "role": {"id": 1}
                 }}
+        return self.tempo_teams_add_member_raw(team_id, member_data=data)
+
+    def tempo_teams_add_membership(self, team_id, member_id):
+        """
+        Add team member
+        :param team_id:
+        :param member_id:
+        :return:
+        """
+        data = {"teamMemberId": member_id,
+                "teamId": team_id,
+                "availability": "100",
+                "role": {"id": 1}
+                }
+        url = "rest/tempo-teams/2/team/{}/member/{}/membership".format(team_id, member_id)
+        return self.post(url, data=data)
+
+    def tempo_teams_add_member_raw(self, team_id, member_data):
+        """
+        Add team member
+        :param team_id:
+        :param member_data:
+        :return:
+        """
+        url = 'rest/tempo-teams/2/team/{}/member/'.format(team_id)
+        data = member_data
         return self.post(url, data=data)
 
     def tempo_teams_get_members(self, team_id):
@@ -1896,7 +2044,7 @@ class Jira(AtlassianRestAPI):
         if limit:
             params['maxResults'] = limit
         if state:
-            params['state'] = [state]
+            params['state'] = state
         url = 'rest/agile/1.0/board/{boardId}/sprint'.format(boardId=board_id)
         return self.get(url, params=params)
 
