@@ -1,10 +1,12 @@
 # coding=utf-8
 import json
 import logging
-from six.moves.urllib.parse import urlencode
+
 import requests
 from oauthlib.oauth1 import SIGNATURE_RSA
 from requests_oauthlib import OAuth1
+from six.moves.urllib.parse import urlencode
+
 from atlassian.request_utils import get_default_logger
 
 log = get_default_logger(__name__)
@@ -20,9 +22,10 @@ class AtlassianRestAPI(object):
     response = None
 
     def __init__(self, url, username=None, password=None, timeout=60, api_root='rest/api', api_version='latest',
-                 verify_ssl=True, session=None, oauth=None, cookies=None, advanced_mode=None):
-
-        if ('atlassin.net' in url or 'jira.com' in url) and '/wiki' not in url:
+                 verify_ssl=True, session=None, oauth=None, cookies=None, advanced_mode=None, kerberos=None):
+        if ('atlassian.net' in url or 'jira.com' in url) \
+                and '/wiki' not in url \
+                and self.__class__.__name__ in 'Confluence':
             url = self.url_joiner(url, '/wiki')
         self.url = url
         self.username = username
@@ -41,10 +44,28 @@ class AtlassianRestAPI(object):
             self._create_basic_session(username, password)
         elif oauth is not None:
             self._create_oauth_session(oauth)
+        elif kerberos is not None:
+            self._create_kerberos_session(kerberos)
 
     def _create_basic_session(self, username, password):
         self._session.auth = (username, password)
 
+    def _create_kerberos_session(self, kerberos_service):
+        try:
+            import kerberos as kerb
+        except ImportError as e:
+            log.debug(e)
+            try:
+                import kerberos_sspi as kerb
+            except ImportError:
+                raise ImportError("No kerberos implementation available")
+        __, krb_context = kerb.authGSSClientInit(kerberos_service)
+        kerb.authGSSClientStep(krb_context, "")
+        auth_header = ("Negotiate " + kerb.authGSSClientResponse(krb_context))
+        self._update_header("Authorization", auth_header)
+        response = self._session.get(self.url, verify=self.verify_ssl)
+        response.raise_for_status()
+
     def _create_oauth_session(self, oauth_dict):
         oauth = OAuth1(oauth_dict['consumer_key'],
                        rsa_key=oauth_dict['key_cert'], signature_method=SIGNATURE_RSA,
@@ -125,6 +146,7 @@ class AtlassianRestAPI(object):
             verify=self.verify_ssl,
             files=files
         )
+        response.encoding = 'utf-8'
         if self.advanced_mode:
             self.response = response
             return response
@@ -178,18 +200,20 @@ class AtlassianRestAPI(object):
         :param trailing: OPTIONAL: for wrap slash symbol in the end of string
         :return:
         """
-        answer = self.request('GET', path=path, flags=flags, params=params, data=data, headers=headers,
+        response = self.request('GET', path=path, flags=flags, params=params, data=data, headers=headers,
                               trailing=trailing)
+        if self.advanced_mode:
+            return response
         if not_json_response:
-            return answer.content
+            return response.content
         else:
-            if not answer.text:
+            if not response.text:
                 return None
             try:
-                return answer.json()
+                return response.json()
             except Exception as e:
                 log.error(e)
-                return answer.text
+                return response.text
 
     def post(self, path, data=None, headers=None, files=None, params=None, trailing=None):
         response = self.request('POST', path=path, data=data, headers=headers, files=files, params=params,
@@ -220,4 +244,11 @@ class AtlassianRestAPI(object):
         :return: Empty dictionary to have consistent interface.
         Some of Atlassian REST resources don't return any content.
         """
-        self.request('DELETE', path=path, data=data, headers=headers, params=params, trailing=trailing)
+        response = self.request('DELETE', path=path, data=data, headers=headers, params=params, trailing=trailing)
+        if self.advanced_mode:
+            return response
+        try:
+            return response.json()
+        except ValueError:
+            log.debug('Received response with no content')
+            return None
